#ifndef MODELING_OP_H
#define MODELING_OP_H 1


#include "gVec.h"
#include "op.h"
#include "hypercube.h"
#include "cubeData.hpp"
#include "cubeAlgo.h"
#include "dmRsf.hpp"
#include "modelingFD_aux.h"

namespace Yang {


//Aux function
void calcTimeParamsFor2wayModeling(FDModelingTimeParams* timePar, float maxFreq, float vmax, float vmin, const hypercube &modelGridShape);

   class vel;

//Forward declaration of template class
//template <typename T>
//class dmRsf_oc;

   
class modelingOp: public op  //Define the class of wave propagation operator
{
public:
   action_type mAction;
   FDModelingParams* m_pars; //all modeling params.
   
public://methods
   modelingOp():op(),m_pars(NULL) {
   }
   
   bool isModeling() {
      return (mAction==acModeling);
   }
   
   //Maybe should change it to born-modeling and adj born-modeling
   bool isRTM() {
      return (mAction==acMigration);
   }
   bool isAdjRTM() {
      return (mAction==acBornModeling);
   }
   
   //init the operator using the wave propagation params. e.g. dtComp
   virtual void initOperator(FDModelingParams* params, action_type ac_type) = 0;
   virtual void initOperatorEx(FDModelingParams* params, action_type ac_type) = 0;
   virtual void modeling_op(cubeData<float>* pVelCube, dmRsf<float>* pRsfSeisRec) = 0;
   virtual void bornModeling_op(cubeData<float>* pReflCube, dmRsf<float>* pRsfSeisRec) = 0;
   virtual void migration_op(dmRsf<float>* pRsfSeisRec, cubeData<float>* pImageCube) = 0;
   virtual void migration_op(const string& seisTag, cubeData<float>* pImageCube) {}
   virtual void TomAdj_op (cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS) = 0;
   virtual void TomFwd_op (cubeData<float>* pDeltaS, cubeData<float>* pDeltaI) = 0;
};

//declaration
 class SR_managerFD;      //Src/Recv Information manager, collect the info of src/receiver geometry, src wavelet and how to exert source
 class boundaryFD;  //how to propagate wavefield by one step; which consists of FDscheme, wavefiled damping and absorption at the boundary.
 class paddingFD;

//General Finite difference modeling operator
class modelingFDgen_op : public modelingOp 
{
public:
   //data memter
   //declare friend class which are the component of the modelingFD operator
   friend class SR_managerFD;
   friend class boundaryFD;
   friend class paddingFD;

   
   SR_managerFD * m_pSR_manager;
   boundaryFD  *m_pBoundaryFD;
   paddingFD   *m_pPaddingFD;

   stdvecInt   mOffsetPandModel; //becoz shape of wavefield is bigger than shape of vel model by 2*stencilWid
protected:
   enum temporal_accuracy
   {
      order2, order4
   };
   temporal_accuracy m_t_accuracy;
   
   impose_receiver_record_type mImposeRecvType;
      
   
   cubeData<float>  mS0cube; //hold initial slowness per shot
   cubeData<float>  mReflCubePerShot; //ReflModel aligned to mShapeWavfield, inner object
   cubeData<float>  mReflOffCubePerShot; //ReflModel prestack , inner object
   //temporary memories for every shot
   cubeData<float>  mRTM_imageCubePerShot, mRTM_imageCubePerShot2;
   cubeData<float>  mSrcIntensePerShot, mRecvIntensePerShot;
   //For prestack imaging I(x,h)=p1(x-h)*p2(x+h)
   cubeData<float>  mRTM_offsetImgCubePerShot, mRTM_offsetImgCubePerShot2;
   
   
protected:
      bool m_bAssembled;
public:
      modelingFDgen_op(): modelingOp(), m_pSR_manager(NULL), m_pBoundaryFD(NULL),
      m_pPaddingFD(NULL), 
      m_bAssembled(false) {} 
   
   //Assemble all the component like SR_manager and boundary, etc. 
   void assemble(SR_managerFD * pSR, boundaryFD* pBC, paddingFD* pPD) {
       m_pSR_manager = pSR; m_pBoundaryFD = pBC;  m_pPaddingFD = pPD;
	   m_bAssembled = true;
   }
   virtual ~modelingFDgen_op(){
   }
   
   //called before each invoking each _op func
   void checkInOutSize();
   virtual void initOperatorEx(FDModelingParams* params, action_type ac_type);
   void unInitOperatorEx() {}
   
   virtual void modeling() {VERIFY(0);}
   virtual void modeling_op(cubeData<float>* pVelCube, dmRsf<float>* pRsfSeisRec);
   virtual void RTM() = 0;
   virtual void migration_op(dmRsf<float>* pRsfSeisRec, cubeData<float>* pImageCube);
   virtual void migration_op(const string& dataTag, cubeData<float>* pImageCube) {}
   virtual void adjRTM() = 0;
   virtual void bornModeling_op(cubeData<float>* pReflCube, dmRsf<float>* pRsfSeisRec);

   virtual void TomAdj();
   virtual void TomAdj_op (cubeData<float>* pDeltaIh, cubeData<float>* pDeltaS);
   virtual void TomFwd();
   virtual void TomFwd_op(cubeData<float>* pDeltaS, cubeData<float>* pDeltaI);

   virtual void initPerShot(int ishot);
   
protected:
   virtual void initOperator(FDModelingParams* params, action_type ac_type);
   virtual void unInitOperator() {}
   
   void genPerShotImage(int ishot);
   //Slices could be 2D or 3D
   void alignCropToSliceShapeGrid(cubeData<float>& outReflCube, cubeData<float>& inReflModel, hypercube& sliceShape);
   void cropWfldToVelGrid(cubeData<float>& velOutCube, cubeData<float>& wfldInCube);
   
   
   void getSnapshotShapePerShot(hypercube& shapeSnapshot);
   void getShotAxisForSnap(axis& ax_shot);
   bool shouldRecordSnap(int ishot, int* iIndexOnAxis=NULL);
   void setWavfieldShape(int iShot);
};

//CPU FD modeling
class modelingFD_op: public modelingFDgen_op {
   
protected:
   stdVecDb  stencil;  //FD spatial stencil

   wavFieldGroup   wfGp0; //wavefield group 1, default
   wavFieldGroup   wfGp1; //for the adjRTM modeling
   cubeData<ETYPE>      dpDdx2_cb; //cube used for storing 2nd order in space for wfld
   auto_ptr<cubeData<ETYPE> > m_pWavfieldCheckptPerShot;
   
   //For a slice of src wavefield, these three are per shot
   cubeData<float>   mSrcWavfieldCube;
   cubeData<ETYPE>   mCorrCube;
   
   //Just for Db->float transition.not necessary
   cubeData<float>    mSnapCube;
   

public:
      modelingFD_op(): modelingFDgen_op() {}
	  virtual ~modelingFD_op(){
	  }
   
   //Get the parameters for wave propagation operator. e.g. dt,dx,freq(center freq of wavlet),
   virtual void initOperator(FDModelingParams* params, action_type ac_type);
   virtual void unInitOperator();   
   
   virtual void modeling();
   virtual void RTM();
   virtual void adjRTM();
   //virtual void modeling_op(cubeData<float>* pVelCube, dmRsf<float>* pRsfSeisRec);
   //virtual void migration_op (dmRsf<float>* pRsfSeisRec, cubeData<float>* pImageCube);

   virtual void initPerShot(int iShot);
   void modelingPerShot(int ishot);
   void adjRTM_perShot(int ishot);

   void unInit();   
   void imposeRecv(wavFieldGroup& wfg);
   void takeRecvWavfieldSnapshot(int ishot, int it, cubeData<ETYPE>& wfd, dmRsf<float>* pRecvSnapShot);
   void takeWavCheckpt(int ishot, int it, wavFieldGroup&);
   void correlateSrcRecv(int iShot, int it, const cubeData<ETYPE>& recvwfld);

protected:
   bool getWindowInSnapshot(windParam& window, int ishot, int it);
	bool getWindowInWavfieldCheckpt(windParam& window, int it);   
   void fwdSrcPropagation(int iShot);
   void setStencil(int _stenWid);

   virtual void propagate(const cubeData<ETYPE>& p0cube,const cubeData<ETYPE>& p1cube,cubeData<ETYPE>& p2cube);
   virtual void propagate_2t(const cubeData<ETYPE>& p0cube,const cubeData<ETYPE>& p1cube,cubeData<ETYPE>& p2cube) = 0;
   virtual void propagate_4t(const cubeData<ETYPE>& p0cube,const cubeData<ETYPE>& p1cube,cubeData<ETYPE>& p2cube) = 0; //4th order in time
   virtual void calcSpatialDeriv( cubeData<ETYPE> *in, cubeData<ETYPE> * out) = 0;

   void takeSnapshot(int ishot, int it, cubeData<ETYPE>& wavfield, dmRsf<float>* pSnapShot);
   virtual void corrWithOffset(cubeData<float>& offCubeOut, const cubeData<float>& wf1,const cubeData<float>& wf2) = 0;
};

} //end of namespace

#endif

