#include "stdafx.h"
#include "solverNonlinearSearch.h"
#include "cubeData.hpp"
#include "commonOps.h"
#include "misc.h"
#include "cubeAlgo2.h"

// TODO: your implementation here
using namespace Yang;

//given (x,y) 3 pts, fit parabola ax^2+b^x+c
void fitParabola(float coef[3], float x[3], float y[3])
{
   float x1,x2,x3;
   float y1,y2,y3;
   x1 = x[0]; x2=x[1]; x3=x[2];
   y1 = y[0]; y2=y[1]; y3=y[2];
   
   float denom = (x1-x2)*(x1-x3)*(x2-x3);
   coef[0] = (x3*(y2-y1) + x2*(y1-y3) + x1*(y3-y2)) / denom;
   coef[1] = (x3*x3*(y1-y2) + x2*x2*(y3-y1) + x1*x1*(y2-y3)) / denom;
   coef[2] = (x2*x3*(x2-x3)*y1 + x3*x1*(x3-x1)*y2 + x1*x2*(x1-x2)*y3)/denom;
}

// Constructor implementation
solverNonlinearSearch::solverNonlinearSearch(
  objFunc* pObj, const gVecComponent* pInitModel, solverParam* pSolParam)
:m_pObjFunc(pObj), m_pOpSmooth(NULL), m_pSolverPar(pSolParam)
{
   m_apCurModel = auto_ptr<gVecComponent>( new gVecComponent());
   m_apTrialModel = auto_ptr<gVecComponent>( new gVecComponent());
   //Init the cubeData ptr
   pInitModel->copyTo(*m_apCurModel);
   pInitModel->allocMemoryTo(*m_apTrialModel);
   
   if (m_pSolverPar->bSmoothSearchDir) {
     //create the smoothing operator
     m_pOpSmooth = new opSpline();
     const hypercube& datShape = pInitModel->domain();
     this->setSmoothParam(pSolParam->nSpl);
   
     //Set the scaling quantities, if we change the smoothing over iterations.
     stdVecInt ns_model = datShape.get_ns();
     for (int i=0; i<3; i++) {
       m_pSolverPar->fsSpl[i] = 1.0f;
     }
     assert(m_pSolverPar->maxIter != 0);
     for (size_t i=0; i<ns_model.size(); i++) {
       if (m_pSolverPar->nSpl[i]!=0) {
         if (i != 0) {
           m_pSolverPar->fsSpl[i] = powf(0.25f*float(ns_model[i])/float(m_pSolverPar->nSpl[i]),1.0f/m_pSolverPar->maxIter); //Make the spline (finest is 4 grid points), i.e increase smoothing.
         } else { //z direction, the smooth can be weaker
           m_pSolverPar->fsSpl[i] = powf(0.5f*float(ns_model[i])/float(m_pSolverPar->nSpl[i]),1.0f/m_pSolverPar->maxIter); //Make the spline finer along iterations. (finest is 2 grid points), i.e increase smoothing.
         }
       }
     }
     PRVEC(m_pSolverPar->fsSpl,3);
   }
   PRV(m_pSolverPar->errTol);
   
   //Check velocity gradient
   if (m_pSolverPar->bMaskGradient) {
     VERIFY(m_pSolverPar->gradMask.domain().sameSize(m_apCurModel->domain()));
   }
}   

// Destructor implementation
solverNonlinearSearch::~solverNonlinearSearch() {
  if (m_pOpSmooth) {
    delete m_pOpSmooth;
  }
}

//Put in the smooth operator
void solverNonlinearSearch::setSmoothParam(int nSpl[3]) {
  FUNC_TRACE_BEG(solverNonlinearSearch::setSmoothParam);
   assert(m_pSolverPar->bSmoothSearchDir);
   //Get the model space shape
   const hypercube& modelSpaceShape = m_apCurModel->domain();
   int ndim = modelSpaceShape.get_ndim();
   assert(ndim <= 3);
   //now create the spline model shape, only ns matters
   vecAxis axes; axis ax,ax_d; ax.o = 0.0f;
   for (int idim=0; idim<ndim; idim++) {
     ax.n = nSpl[idim];
     axes.push_back(ax);
   }
   hypercube splModShape(axes);
   m_pOpSmooth->init(splModShape,modelSpaceShape);
   
  FUNC_TRACE_END(); 
}


//Maybe change cubeData into a gVec is better
//return true if there is improvement
bool solverNonlinearSearch::updateModel(const gVecComponent& searchDir, float maxPerturbScale)
{
   bool ret = true;
	
   VERIFY(searchDir.sameSizeVec(m_apCurModel.get()));

   //find the scale for search Dir first.
   float maxval_searchDir = searchDir.maxvalAbs();
   
   float fmax,fmin,fClipRat; //allowed slowness range
   fmax = m_pSolverPar->fmaxValAllow;
   fmin = m_pSolverPar->fminValAllow;
   //PRV(fmax); PRV(fmin);
   int i;
   float stepsizes[3], objFuncVals[3], coefParabola[3];
   float stepRef;
   //the stepsize is 1/2,1,2 times of the allowed model perturbation given maxPerturbScale
   if (equalf(m_RMS_initModel,0.0f)) {
     stepRef = 1.0f; //Just set 1.0f, if does not know the what the scaling should be.
   } else {
     stepRef = (maxPerturbScale*m_RMS_initModel)/maxval_searchDir;
   }
   
   //multiply stepRef LATER
   stepsizes[0] = 0.0;
   stepsizes[1] = 0.1;
   stepsizes[2] = 1.0;
   
   //perturb 2 step size respectively and evaluate the objective functions.
   //build Trial model
   objFuncVals[0] = m_curObjFuncValue;
   for (i=1; i<3; i++)
   {
	  m_apTrialModel->scaleAddVec(m_apCurModel.get(),&searchDir,stepsizes[i]*stepRef);
    //Only this is not enough, need to clip the values in the new model
    fClipRat = clipCubeVal(*m_apTrialModel,fmax,fmin,true,true); 
    if (fClipRat>0.01f) {
      prt("\nClip the trial solution by ratio %f",fClipRat);
    }
	  //evaluate obj function
	  //m_apTrialModel->dump("dump4");
	  objFuncVals[i] = m_pObjFunc->evaluate(*m_apTrialModel);
   }
   
   //optimal among current 3 pts
   float opt3ObjFunc; float opt3Stepsize;
   stdVecFloat vecfval(objFuncVals, objFuncVals+3);
   stdVecFloat vecfstep(stepsizes, stepsizes+3);
   
   stdVecFloat::iterator minIter = std::min_element(vecfval.begin(),vecfval.end());
   int optIndex = minIter - vecfval.begin();
   opt3ObjFunc = vecfval[optIndex];
   opt3Stepsize = vecfstep[optIndex];
   
   //Then fit a parabola among these 3 pts.
   fitParabola(coefParabola,stepsizes,objFuncVals);
   
   float optObjFunc, optStepsize;
   float optParaStepsize = -1.0f, optParaObjFunc = -1.0f;
   //decide the optimal point, according to the shape of the parabola
   //First sets the optimum as the current best.
   optObjFunc = opt3ObjFunc;
   optStepsize = opt3Stepsize;
   
   //if the parabola is upside-down or just a line
   if (coefParabola[0] < 0.0f || equalf(coefParabola[0],0.0f)) {
	  verbPrt("pathological parabola! curvature %f \n", coefParabola[0]);
   }
   else //bottom-up regular parabola
   {
	  //Get the optimal pt, -b/2a
	  optParaStepsize = -coefParabola[1]/(2.0f*coefParabola[0]);
	  
	  //see the relative postion with respect to the 3 anchor pts on the parabola
	  
	  //Test the obj func at this pt
	  m_apTrialModel->scaleAddVec(m_apCurModel.get(),&searchDir,optParaStepsize*stepRef);
    fClipRat = clipCubeVal(*m_apTrialModel,fmax,fmin,true,true);
    //evaulate the obj function at that optimal pt
	  optParaObjFunc = m_pObjFunc->evaluate(*m_apTrialModel);
	  
	  //if parabola bottom is even better, BE AWARE THAT optParaObjFunc could be NAN !
	  if (optParaObjFunc < opt3ObjFunc)
	  {
		 optObjFunc = optParaObjFunc;
		 optStepsize = optParaStepsize;
	  }
   }
	
	verbPrt("\nstep size: %f,%f,%f,%f, objFunc=%f,%f,%f,%f optimal size %f, funcVal %f\n", stepsizes[0],stepsizes[1],stepsizes[2],optParaStepsize,
			  objFuncVals[0],objFuncVals[1],objFuncVals[2], optParaObjFunc, optStepsize, optObjFunc );
	
   //select a best point among these tested locations
	//if there is improvements
	if (!equalf(optStepsize,0.0f)) {
		if (optStepsize < 0.0f) {
		  verbPrt("\noptimum step size is negative !! %f \n", optStepsize);
		}
		
		//update the model using this point
		m_curObjFuncValue = optObjFunc;
		m_apTrialModel->scaleAddVec(m_apCurModel.get(),&searchDir,optStepsize*stepRef);
    fClipRat = clipCubeVal(*m_apTrialModel,fmax,fmin,true,true);
		//switch the ptrs
		std::swap(m_apTrialModel,m_apCurModel);
	}
	else //there is no improvement, search stalls
	{
		ret = false;
	}
   return ret;
}

//The non-linear line search optimization work flow
//Given the constrain of the maximum iterations avaible and the relative error difference between iterations
void solverNonlinearSearch::solve( gVecComponent& finalModel, stdVecFloat& finalErrorVec, int& finalIter, std::vector<gVecComponent*>* pVecModels,std::vector<gVecComponent*>* pVecGrads, vector<gVecComponent*>* pVecDatas)
{
  FUNC_TRACE_BEG(solverNonlinearSearch::solve);
  bool bRecordModels =  (pVecModels != NULL);
  bool bRecordGrads  =  (pVecGrads  != NULL);
  bool bRecordDatas  =  (pVecDatas  != NULL);
  
  const solverParam* par = m_pSolverPar;
  float maxPerturbScale = m_pSolverPar->fMaxPerturbScale;
  int maxIter = m_pSolverPar->maxIter;
  float errorAllowed = m_pSolverPar->errTol;
  
  //Init the output params
  finalIter = 0;
  finalErrorVec.clear();
   
  //Evaluate the initial model
  objFunc* pObjFunc = this->m_pObjFunc;

  //evaluate the initial model
  float fInitError = pObjFunc->evaluate(*m_apCurModel);
  finalErrorVec.push_back(fInitError);
  m_curObjFuncValue = fInitError;
  m_RMS_initModel = m_apCurModel->RMS_NZ(1e-6);
  PRV(m_RMS_initModel);
  
  verbPrt("\nfinish initial evalution of obj func");
  
  //cubeData<float>& dCalInit = pObjFunc->getCalcData();
  //dCalInit.dump("dump6");
  
  int iter;
  float fObjFuncOld, fObjFuncNew;
  gVecComponent searchDir, searchDir2;
  this->m_apCurModel->allocMemoryToVec(& searchDir);
  this->m_apCurModel->allocMemoryToVec(& searchDir2);
  for (iter=0; iter<maxIter; iter++) {
    
    if (iter !=0 ) {
      //needs to reevaluate obj in order to calculate gradient
      pObjFunc->evaluate(*m_apCurModel);
      pObjFunc->getCalcData().dump("fImage");
      pObjFunc->callBackAtIter(iter); //Call back func for objFunc to update info
    }
    
    if (bRecordDatas) {
      gVecComponent* pInterModel = new gVecComponent();
      m_pObjFunc->getCalcData().copyTo(*pInterModel);
      pVecDatas->push_back(pInterModel);
      dumpVecModels(*pVecDatas,"dats");
    }

    //calculate the update direction
    pObjFunc->calcSearchDir(searchDir);
    //Mask the gradient if necessary
    if (m_pSolverPar->bMaskGradient) {
      searchDir.multiply(m_pSolverPar->gradMask);
    }
    //Might need smooth the gradient
    searchDir.copyToVec(&searchDir2);
    if (m_pSolverPar->bSmoothSearchDir) {
      this->smoothGradient(searchDir2, searchDir);
      searchDir2.dump("dump10");  searchDir.dump("dump11"); LOGC(dump the search directions);
      //m_pOpSmooth->getModShape().info("\nspl modshape"); m_pOpSmooth->getDatShape().info("\nspl datshape");
    }
    
    if (bRecordGrads) {
      gVecComponent* pInterModel = new gVecComponent();
      //First store the original gradient
      searchDir2.copyToVec(pInterModel);
      pVecGrads->push_back(pInterModel);
      //Then store the smoothed gradient
      gVecComponent* pInterModelsm = new gVecComponent();
      searchDir.copyToVec(pInterModelsm);
      pVecGrads->push_back(pInterModelsm);
      dumpVecModels(*pVecGrads,"grads");
    }

    //searchDir2.dump("dump5"); searchDir.dump("dump6");
    //searchDir.dump("dump7");
    fObjFuncOld = this->m_curObjFuncValue;
 
    verbPrt("\n at iter = %d,", iter);   
    //update the model
    bool bImprove = this->updateModel(searchDir,maxPerturbScale);
    
    //Might need to adjust the smoothing operator and the maxStepLength
    //The formula is stepLenMax = (base1 + (1-base1)*(fdecay^(iter)))*fMaxPerturbScale
    if (par->bDecreaseMaxStepLengthWithIteration) {
      maxPerturbScale = (par->fbase1+(1.0f-par->fbase1)*powf(par->fdecay,float(iter+1)))*par->fMaxPerturbScale;
      PRV(maxPerturbScale);
    }
    if (par->bSmoothSearchDir && par->bChangeSmoothParamWithIteration){
      //Need to refine the spline points.
      int nspls[3] = {0,0,0};
      stdVecInt ns_model = m_apCurModel->domain().get_ns();
      stdVecInt ns_curSpl = m_pOpSmooth->getModShape().get_ns();
      for (size_t i=0; i<ns_curSpl.size(); i++) {
        nspls[i] = MIN(ns_model[i],int(ns_curSpl[i]*par->fsSpl[i]));
      }
      this->setSmoothParam(nspls);
    }
      
    fObjFuncNew = this->m_curObjFuncValue;
    finalErrorVec.push_back(fObjFuncNew);
    
    //Now record the updated model
    if (bRecordModels) {
      gVecComponent* pInterModel = new gVecComponent();
      m_apCurModel->copyToVec(pInterModel);
      pVecModels->push_back(pInterModel);
      //Dump the models and grads.
      dumpVecModels(*pVecModels,"mods");
    }    
    
    if (!bImprove) {
     verbPrt("\nNo improvement in nonlinear solver!!!\n");
     break;
    }

    //see how much the error has improved.
    VERIFY(fObjFuncNew <= fObjFuncOld);
    if ( fabs(fObjFuncNew) < fabs(fInitError)*errorAllowed) {
      prt("\nThe current Obj func %f is small enough in terms of Initial Error %f, quit the optimization loop at iter=%d \n",
           fabs(fObjFuncNew), fabs(fInitError), iter);
      break;
    }
    if ( fabs(fObjFuncOld - fObjFuncNew) < fabs(fInitError) * errorAllowed) {
      verbPrt("\ndifference %f is small enough in terms of Initial Error %f, quit the optimization loop at iter=%d \n",
              fabs(fObjFuncOld - fObjFuncNew), fabs(fInitError), iter);
      break;
    }
    if ( fabs(fObjFuncOld - fObjFuncNew) < fabs(fObjFuncNew) *errorAllowed) {
      verbPrt("\ndifference %f is small enough in terms of current Error %f, quit the optimization loop at iter=%d \n",
               fabs(fObjFuncOld - fObjFuncNew), fabs(fObjFuncNew), iter);
      break;
    }
  } //End for iter
  
  finalIter = iter;
  this->m_apCurModel->copyToVec(&finalModel);
  
  //report statistics
  prt("nIter=%d, and the final Error Vec is: \n",finalIter);
  prtStdVec(finalErrorVec);
  
  
  //Again evaluate the final model, if you don't need the final datat, i.e F(m_final) then it is not necessary.
  pObjFunc->evaluate(*m_apCurModel);
  //cubeData<float>& dCalFinal = pObjFunc->getCalcData();
  //dCalFinal.dump("finalData");
  FUNC_TRACE_END();
}

void solverNonlinearSearch::smoothGradient(const gVecComponent& inputSearchDir, gVecComponent& searchDir) {
  //assert the size match as well.
  if (m_pSolverPar->eSmooth == solverParam::sm_spline) {
    CONST_CAST_REF(input,inputSearchDir,gVecComponent);
    m_pOpSmooth->smooth(input,searchDir,m_pSolverPar->repeat);
  } else if (m_pSolverPar->eSmooth == solverParam::sm_rect) {
    char szPar[256];
    sprintf(szPar," repeat=%d rect1=%d rect2=%d ", m_pSolverPar->repeat, m_pSolverPar->rect[0], m_pSolverPar->rect[1]);
    seplib::callSepTool(searchDir, inputSearchDir, "Smooth", szPar, false);
  } else {
    VERIFY(0);
  }
}
