#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "modelingFD2d_op.h"
#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"

#include "param_entry.h"
#include "myOMP.h"
#include "../misc.h"

/**
 * Main application entry point 
 */
 
using namespace Yang;
using namespace std;

//Currently quadratic form, theta= u*(offset)^2
//assume input is 1d, axis is along offset
void setDthetaOverDu(cubeData<float>& cube)
{
   axis ax= cube.domain().get_axis(1);
   VERIFY(cube.domain().get_ndim()==1);
   int n = ax.n;
   float o = ax.o;
   float d = ax.d;
   float x;
   for(int i=0;i<n;i++)
   {
	  x = o + i*d;
	  cube[i] = x*x;
   }
}

void modelAndCalcCorrPanel(cubeData<float>& C_cube,int iStartLag,int iEndLag,cubeData<float>& d_calCube, const cubeData<float>& d_obsCube, modelingFD_op* pModeling, vel* pVelCur)
{
   //simulate the calc data from the current slowness s
   prt("simulate data from the current slowness s");
   pModeling->setVelocity(pVelCur);
   pModeling->modeling();
   //extract the seismic data of modeling
   dmRsf<float>  *pDataCalc = pModeling->m_pSR_manager->m_pSeisRec;
   pDataCalc->resetCounter();
   pDataCalc->grabAll(&d_calCube);

   //along with observed data, calculate the correlation panel   
   //After getting the calculated data P_cal, calculate the cross-correlation with P_obs.
   crossCorr1d(C_cube,d_calCube,d_obsCube,iStartLag,iEndLag);
}

{
   //After modeling, save the snapshot and seisrecord
   m_pars->pSrcSnapshot->save(); //Since modeling, need to save wavefield snapshot to file
   m_pars->RecvSnapshot->save(); //Save second wavefield to file


}
//Doing beam wave-equation tomography
int main (int argc, char *argv[])
{
   //Init openMP, use 7 cores
   init_omp(1);

   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   bool bInput = true;
   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,bInput));

   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,bInput));

   //Read in wavelet and initial velocity, just input a Guassian wavelet, beacause it will take 2nd 
   //derivate to create the ricker wavelet.
   waveletObj->loadWavelet();
   velObj->loadVel();

   //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
   auto_ptr<modelingFD2d_op> modelingObj(new modelingFD2d_op(),acModeling);
   //Create the padding and boundary processor
   auto_ptr<SR_managerFD2d>     SR_managerObj(new SR_managerFD2d(csz_tag_seisRec,modelingObj.get(),waveletObj.get()));
   
   

   //chain two padding method together
   float dampingExtent = 0.5f;
   auto_ptr<paddingFD2dDamping> paddingDampingObj ( new paddingFD2dDamping(modelingObj.get(),NULL,dampingExtent));
   auto_ptr<paddingFD2dSponge1> paddingObj (new paddingFD2dSponge1(modelingObj.get(), paddingDampingObj.get()));   
   //boundary condition
   auto_ptr<boundaryFD2dAbsorb> boundaryObj(new boundaryFD2dAbsorb(modelingObj.get()));

   //assemble
   modelingObj->assemble(SR_managerObj.get(), boundaryObj.get(),paddingObj.get(),velObj.get());

   //Create another RTM imaging operator to calculate the RTM
   auto_ptr<modelingFD2d_op> modelingObj(new modelingFD2d_op(),acRTM);
   
   //Read in the parameters for the modelingOperator
   fetchParamAllModelingFD(modelingObj.get(),waveletObj.get(),pars);
   initAllModelingFD(modelingObj.get(),waveletObj.get());   
   prt("finish the initialization of the modeling Operator.\n");

   //Load the corrlation window
   int iStartLag, iEndLag;
   iStartLag = pars->get_int(csz_corrStartLag);
   iEndLag = pars->get_int(csz_corrEndLag);
   

   
   //load the observed data
   dmRsf<float> *pDataObs;
   //currently Do incore
   pDataObs = new cubeData(csz_dataObs,dm::IN,"data_obs");
   pDataObs->init(NULL);
   cubeData<float> d_obsCube;
   cubeData<float> pDataObs->seq_grabAll(d_obsCube);
   

   vel2d *pVelCur = velObj.get();
   cubeData<float> C_cube,d_calCube;
   
   //Here we start to calculate the gradient of Jlocal
   modelAndCalcCorrPanel(C_cube,iStartLag,iEndLag,d_calCube,d_obsCube,modelingObj.get(),pVelCur);
   
   //take the derivative of C'(tau;s0), and keep also C(tau;s0)
   cubeData<float> C1_cube,C2_cube;
   C1_cube.alloc(C_cube.domain());
   C2_cube.alloc(C_cube.domain());   
   timeDeriv1st(C1_cube,C_cube);
   timeDeriv2nd(C2_cube,C_cube);
   
   int iShot,nShots,iGroup,nGroup;
   int subGroupSize = pars->get_int(csz_subGroupSize); //e.g. beam size;
   
   //Assert receiver size match
   int nRecv = C_cube.domain().get_n(2);
   nGroup = nRecv / subGroupSize;
   VERIFY(nRecv%subGroupSize==0);

   
   
   cubeData<float> C_subCube,C1_subCube,C2_subCube;
   cubeData<float> trcCube,spreadCube;

   cubeData<float> dThetaOverDu,dOffsetArray2; //Use to save array like D(theta)/D(u)

   cubeData<float> d_injectCubePerShot,d_injectCubePerShot2;
   cubeData<float> d_injectCube;
   cubeData<float> cubeVelGradient;
   
   //axis for local offset
   int n; float o,d;
   d = C_cube.domain().get_d(2);
   n = subGroupSize;
   //place the origin(0 offset) at the hyperbola's center
   o = (subGroupSize/2)*(-d);
   axis ax_offset(n,o,d,"local offset","m");   
   hypercube offsetShape(vector<axis>(1,ax_offset));
   vector<axis> axesSubCube;
   axesSubCube.push_back(C_cube.domain().get_axis(1));
   axesSubCube.push_back(ax_offset);
   hypercube subCubeShape(axesSubCube);
   
   dThetaOverDu.alloc(offsetShape);
   dOffsetArray2.alloc(offsetShape);
   
   setDthetaOverDu(dThetaOverDu);
   
   float DJDu, denom, term3;
   
   int dimPerShot = 2;
   cubeData<float> d_obsCubePerShot;
   cubeData<float> C_cubePerShot, C1_cubePerShot, C2_cubePerShot;
   int f,j;
   windParam subGroupWindow,shotWindow;
   subGroupWindow.resize(dimPerShot);
   shotWindow.resize(dimPerShot+1);
   
   //shape is a subcube the big data cube
   hypercube shapeDataPerShot = d_obsCube.domain().return_compatible(dimPerShot);
   hypercube shapeCorrPerShot = C_cube.domain().return_compatible(dimPerShot);
   
   d_injectCube.alloc(C_cube.domain());
   for(iShot=0;iShot< nShots; iShot++)
   {
      d_obsCubePerShot.alloc(shapeDataPerShot);
      C_cubePerShot.alloc(shapeCorrPerShot);
      //aux cubes, remember that finally a correlation cube is injected
      d_injectCubePerShot.alloc(shapeCorrPerShot);
      d_injectCubePerShot2.alloc(shapeCorrPerShot);
      
	  //read in d_obsCubePerShot
      shotWindow[dimPerShot].n = 1;
      shotWindow[dimPerShot].f = iShot;
      
	  d_obsCube.grabXd(&d_obsCubePerShot, shotWindow,dimPerShot);
      C_cube.grabXd(&C_cubePerShot,shotWindow,dimPerShot);
      C1_cube.grabXd(&C1_cubePerShot,shotWindow,dimPerShot);
      C2_cube.grabXd(&C2_cubePerShot,shotWindow,dimPerShot);      
      
      C_subCube.alloc(subCubeShape);
      C1_subCube.alloc(subCubeShape);
      C2_subCube.alloc(subCubeShape);
      
      trcCube.alloc(hypercube(C_subCube.domain().get_axis(1)));
      spreadCube.alloc(C_subCube.domain());
	  //for each subgroup
	  for(iGroup=0; iGroup<nGroup; iGroup++)
	  {

         //subdivide the correlation panel C(tau;s0)
         //Read in C_subCube, C1_subCube, C2_subCube
         subGroupWindow[dimPerShot].n = subGroupSize;
         subGroupWindow[dimPerShot].f = subGroupSize*iGroup;

         C_cubePerShot.grabSeveralXd(&C_subCube,subGroupWindow,1);
         C1_cubePerShot.grabSeveralXd(&C1_subCube,subGroupWindow,1);
         C2_cubePerShot.grabSeveralXd(&C2_subCube,subGroupWindow,1);

		 //sum the traces along the subarray offset
         trcCube.zero();
		 stackTrc(trcCube, C_subCube);
		
         spreadCube.zero();
		 //spread the result back to the entire subarray, Cu(tau;s0)
		 spreadTrc(spreadCube,trcCube);
		 
		 //multiply Cu and C' panel and sum along the tau axis
		 spreadCube.multiply(C1_subCube);
		 sumAlongTimeTrc(dOffsetArray2,spreadCube)
		 
		 //dot product with D(theta)/D(u)
		 float DJDu = dotProd<float>(dOffsetArray1,dOffsetArray2);
		 
		 //Now we have one scaler for DJ/u(x_bar) for each subgroup location x_bar
		 
		 //Calculate the denominator
		 //calculate the 2nd derivative C''(tau;s0)
		 //scale the C''(tau;s0) by D(theta)/D(u) along the local offset.
		 scaleTrc(C2_subCube,dThetaOverDu);
		 //Dot prod the previous and the C(tau;s0)
		 denom = dotProd<float>(C2_subCube,C_subCube);
		 
		 //then we calculate the III term.
		 term3 = DJDu/denom;
		 
		 //scale the C'(tau;s0) by offset using the scaler we just obtained.
		 scaleTrc(C1_subCube, dThetaOverDu);
		 C1_subCube.scale(term3);
		 
		 //Assemble the data together.
		 //put C1_subcube back to the large cube
		 C_cubePerShot.updateSeveralXd(C1_subCube,subGroupWindow,1);
	  }
         
		 //cross-correlation with observed data again, the time direction is not confirmed yet !!!!
		 crossCorr1d(d_injectCubePerShot,C_cubePerShot,d_obsCubePerShot);
		 
		 //back project this data, (take 2nd derivative) because the born-modeling formula
		 timeDeriv2nd(d_injectCubePerShot2,d_injectCubePerShot);

         //Put the data back to a large cube
         d_injectCube.updateXd(d_injectCubePerShot2,shotWindow,dimPerShot)
         //Sum the gradient over all shots.
   } //for each shot
   
   //Do RTM to calculate gradient
   //First set the injected data to the RTM operator
   
   pModeling->RTM();
   //divide by cubic(vel(x)), then we have the search direction.
   //cross-correlate the wavefield with the background calculated wavefield.
   
   cubeVelGradient;
   
   //We use auto_ptr now, no need to delete those ptrs.
   return 0;
}

