#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "modelingFD2d_op.h"
#include "modelingFDgpu.h"

#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"

#include "param_entry.h"
#include "myOMP.h"

#include "FWI_FD.h"
#include "FWI_FD_beamFocus.h"
#include "FWI_FD_DSO.h"
#include "stackTraceC.h"

#include "velAna.param.h"
//#include "../misc.h"

/**
 * For beam focusing, Wave Equation Tomography, gradient 
 * calculation, CUDA enabled
 */


using namespace Yang;
using namespace std;
   
//Doing beam wave-equation tomography for transmission case
/// The input of the program
/// initial velocity model vel
/// observed data data_obs along with its geometry

int main (int argc, char *argv[])
{
   initCallStackTrace();
   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"vel"));
   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
   waveletObj->loadWavelet();
   velObj->loadVel();
   
   //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
   auto_ptr<modelingFDgpu2d_op> modelingObj(new modelingFDgpu2d_op());
   int gpuID = pars->get_int(csz_gpuID,0);
   modelingObj->setGPU_ID(gpuID);
   //Create the padding and boundary processor
   auto_ptr<SR_managerFD2d>  SR_managerObj(new SR_managerFD2d(waveletObj.get()));
   
   //chain two padding method together
   //float dampingExtent = 0.5f;
   //auto_ptr<paddingFD2dDamping> paddingDampingObj ( new paddingFD2dDamping(NULL,dampingExtent));
   auto_ptr<paddingFD2d> paddingObj (new paddingFD2d(NULL));
   //cosDampingFunc dampFunc;
   //paddingDampingObj->m_pDampingFunc = &dampFunc;
		
   //boundary condition
   auto_ptr<boundaryFD2d> boundaryObj(new boundaryFD2d());

   //assemble
   modelingObj->assemble(SR_managerObj.get(), boundaryObj.get(),paddingObj.get());

   //Create another RTM imaging operator to calculate the RTM
   //modelingFD2d_op modelingObj(new modelingFD2d_op(),acMigration);
   
   //Read in the parameters for the modeling Operator
   FDModelingParams params;
   velAnaParam velAnaParams;
   FWI_Param FWIparam;
   //load the observed data
   velAnaParams.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
   velAnaParams.pDataObs->init(NULL);
  velAnaParams.pDataObs ->domain().info("Data obs info");

   float maxFreq,vmax,vmin;
   params.pVel = velObj.get();
   params.pWavelet = waveletObj.get();
   //Read in the peak/center freq of the seismic wavelet
   maxFreq = waveletObj->get_float("peakFreq");
   waveletObj->setPeakFreq(maxFreq);
   vmax = velObj->maxval();
   vmin = velObj->minval();
   
   //The size of whole velocity model
   params.velModelShape = velObj->domain();   
   //set size of mShapeWavfieeld, just simple case, where migration aperture doesnot change
   params.modelGridShape = params.velModelShape;

   //also get the time param from the params
   params.bFreeSurfaceOnTop = pars->get_bool(csz_freeSurface);
   params.stencilWidth = pars->get_int(csz_stenWidth);
   params.bUse4thOrder= pars->get_bool(csz_use4thOrderInTime);
   

   params.ndim = params.velModelShape.get_ndim();
   params.offx = params.stencilWidth;params.offz = params.stencilWidth;
   params.m_nShots 	= pars->get_int(csz_nShots);
   params.mShot1_ix = pars->get_int(csz_shot1_ix);
   params.mShot1_iz = pars->get_int(csz_shot1_iz);
   params.mShotIncX = pars->get_int(csz_shotIncX);
   params.mTrace1ix = pars->get_int(csz_trace1_ix);
   params.mTrace1iz = pars->get_int(csz_trace1_iz);
   params.mTraceIncX = pars->get_int(csz_traceIncX);
   params.m_nTraces = pars->get_int(csz_nTraces);
   params.m_nPad = pars->get_int(csz_npad);

   //Recording Snapshot params
   params.f_shot_snap = pars->get_int(csz_snap_f,0);
   params.j_shot_snap = pars->get_int(csz_snap_j);
   params.n_shot_snap = pars->get_int(csz_snap_n,0);
   checkSnapRecordParams(&params);

   //Read in Absorption type
   int itype = pars->get_int(csz_bdType);
   switch (itype)
   {
      case 1:
         params.m_bdType = abs_ord1;
         break;
      case 2:
         params.m_bdType = abs_ord2;
         break;
      case 3:
         params.m_bdType = abs_Dave;
         break;
      default:
         seperr("!unknown boundary condition type!\n");
   }
   
   //Here I set the data rsf objects
   params.mImposeRecvType = AsForce;
   params.bTake2ndTimeDerivForRTM = pars->get_bool(csz_take2ndTimeDerivRTM,false);
   
   params.pSrcSnapshot 	= new cubeData<float>("",dm::INOUT,"src snapshot");
   params.pRecvSnapshot	= new cubeData<float>("",dm::INOUT,"recv snapshot");

   //The gradient of the local objective function, output
   params.pFinalImage	= new cubeData<float>("",dm::OUT,"");
   params.pImages	= new cubeData<float>("grads",dm::OUT,"");
   
   //calc time parameters
   //Always assume the recording starts from zero time. !! 
   params.timeParam.dtRecord = velAnaParams.pDataObs->domain().get_d(1);
   params.timeParam.totalTime= (velAnaParams.pDataObs->domain().get_n(1)-1) * params.timeParam.dtRecord;

   params.timeParam.dtComp = pars->get_float(csz_dtComp);
   params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
   params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);

   //calc time parameters
   calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.modelGridShape);

   //Load the observed data
   velAnaParams.dObsCube.alloc(velAnaParams.pDataObs->domain());
   velAnaParams.pDataObs->seq_grabAll(&(velAnaParams.dObsCube));
   params.pRsfSeisRec = &velAnaParams.dObsCube;
   
   //Load the corrlation window
   FWIparam.iStartLag = pars->get_int(csz_corrStartLag);
   FWIparam.iEndLag = pars->get_int(csz_corrEndLag);
   velAnaParams.beamSize = pars->get_int(csz_subGroupSize,-1); //e.g. beam size;
   
   int objFuncIndex = pars->get_int("objFunc",1);
   int moveOutFuncType = pars->get_int("moveOutType",1);
   switch(moveOutFuncType) {
   case 1: 
	  velAnaParams.moveOutType = Yang::Linear;
	  break;
   case 2:
	  velAnaParams.moveOutType = Yang::Quadratic;
	  break;
   default:
	  seperr("wrong moveOutType for beam focusing");
	  break;
   }
   
   //Create FWIObj
   float v0,vTrue;
   v0 = pars->get_float("v0",-1);
   vTrue = pars->get_float("vTrue",-1);

   FWI_FD_beamFocusY *p = NULL;
   FWI_FD* pFWIObj = NULL;
   bool bTaperEdge = true; //Has to be true!
   
   switch(objFuncIndex) {
   case 1: 
	  prt("using classical FWI obj Function");
	  FWIparam.mFWItype = FWI_Param::classic; 
	  pFWIObj = new FWI_FD_classic(&velAnaParams,&params,modelingObj.get(),&FWIparam);
	  break;
	  
   case 2: 
	  prt("using YiLuo FWI obj Function");
	  FWIparam.mFWItype = FWI_Param::YiLuo; 
	  pFWIObj = new FWI_FD_YL(&velAnaParams,&params,modelingObj.get(),&FWIparam);
	  break;
   case 22: 
    prt("using YiLuo2 FWI obj Function");
    FWIparam.mFWItype = FWI_Param::YiLuo; 
    pFWIObj = new FWI_FD_YL2(&velAnaParams,&params,modelingObj.get(),&FWIparam);
    break; 
   case 3: 
	  prt("using beam focusing1 FWI obj Function\n");
	  FWIparam.mFWItype = FWI_Param::beamFocusing1;
	  pFWIObj = new FWI_FD_beamFocus1(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 32:
    prt("using beam focusing FWI global obj Function, use only global shift\n");
    FWIparam.mFWItype = FWI_Param::beamFocusing1;
    pFWIObj = new FWI_FD_beamFocus12(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 33:
    prt("using beam focusing, only shift param, new tau^2 obj func");
    FWIparam.mFWItype = FWI_Param::beamFocusing1;
    pFWIObj = new FWI_FD_beamFocus13(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 34:
    prt("using beam focusing, only shift param, maximize stack of local subarray");
    FWIparam.mFWItype = FWI_Param::beamFocusing1;
    pFWIObj = new FWI_FD_beamFocus14(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 35:
    prt("using beam focusing 15, only shift param, use the local ref trace that is close to the center");
    FWIparam.mFWItype = FWI_Param::beamFocusing1;
    pFWIObj = new FWI_FD_beamFocus15(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 4:
	  prt("using beam focusing2 FWI obj Function, slope and curvature\n");
	  FWIparam.mFWItype = FWI_Param::beamFocusing2;
	  pFWIObj = new FWI_FD_beamFocus2(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 40:
    prt("using beam focusing2 FWI obj Function 40,  slope and shift\n");
    FWIparam.mFWItype = FWI_Param::beamFocusing2;
    pFWIObj = new FWI_FD_beamFocus20(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 41:
    prt("using beam focusing2 FWI obj Function, slope and shift, Window no overlap\n");
    FWIparam.mFWItype = FWI_Param::beamFocusing21;
    pFWIObj = new FWI_FD_beamFocus21(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 42:
    prt("using beam focusing2 FWI obj Function, slope and shift, Window overlap \n");
    FWIparam.mFWItype = FWI_Param::beamFocusing21;
    pFWIObj = new FWI_FD_beamFocus22(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 5:
	  prt("using beam focusing Yang's FWI obj Function, one param picking\n");
	  FWIparam.mFWItype = FWI_Param::beamFocusingY1;
	  p = new FWI_FD_beamFocusY1(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
	  p->m_v0 = v0; p->m_vTrue = vTrue;
	  pFWIObj = p; break;
   case 51:
	  prt("using beam focusing Yang's FWI obj Function global shift, one param picking\n");
	  FWIparam.mFWItype = FWI_Param::beamFocusingY1;
	  p = new FWI_FD_beamFocusY1sh(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
	  p->m_v0 = v0; p->m_vTrue = vTrue;
	  pFWIObj = p; break;
   case 6:
	  prt("using beam focusing Yang's FWI obj Function, two param picking\n");
	  FWIparam.mFWItype = FWI_Param::beamFocusingY2;
	  p = new FWI_FD_beamFocusY2(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
	  p->m_v0 = v0; p->m_vTrue = vTrue;
	  pFWIObj = p; break;
   case 7: //DSO1
	  prt("\n Using DSO objective function");
	  FWIparam.mFWItype = FWI_Param::DSO;
	  pFWIObj = new FWI_FD_DSO1(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 8: //DSO2, fake one
	  prt("\n Using DSO objective function!");
	  FWIparam.mFWItype = FWI_Param::DSO;
	  pFWIObj = new FWI_FD_DSO2(&velAnaParams,&params,modelingObj.get(),&FWIparam,bTaperEdge);
    break;
   case 9:  //Yang
     prt("\n Using Yang's objective function!");
     FWIparam.mFWItype = FWI_Param::DSO;
     pFWIObj = new FWI_FD_Yang(&velAnaParams,&params,modelingObj.get(),&FWIparam);
     break; 
   case 10: break;
   default:
	  seperr("wrong  objFunction index");
	  exit(-1);
   }
  //Read Gaussian derivative width
  if (objFuncIndex==9) {
    FWIparam.widthGaussDeriv = pars->get_float("widthGaussDeriv"); //in seconds
  }
  
  read_velAna_param(&velAnaParams, pars);   
   
  //To save time, here I loaded the calculated data
  dmRsf_oc<float> dmCalData("dcal",dm::IN,"calData");
  dmCalData.init(NULL);
  VERIFY(dmCalData.sameSize(velAnaParams.dObsCube));
  FWIparam.dCalCube.alloc(dmCalData.domain());
  dmCalData.seq_grabAll(&FWIparam.dCalCube);
  dmCalData.unInit();
   
  //calcualte the correlation
  pFWIObj->postCalcData();
  FWIparam.dCalCube.dump("dump5");
  velAnaParams.dObsCube.dump("dump6");
  
  if (FWIparam.C_cube.size()!=0) {
   FWIparam.C_cube.dump("corr");
  }
  
  //calculate the bacjProject data
  pFWIObj->calcBackProjData();
  
  FWIparam.backProjCube.dump("backProjData");
  //calculate gradient this time
  velAnaParams.velGradCube.alloc(params.modelGridShape);
  pFWIObj->calcSlowGradient(velAnaParams.velGradCube);
  
  //Dump the src and recv wavefield.   
  params.pSrcSnapshot->unInit();
  params.pRecvSnapshot->unInit();
  
  //Save the gradient to see
  velAnaParams.velGradCube.dump("grad");
  
  //dump RTM image by each shot
  params.pImages->unInit();
  params.pFinalImage->unInit();
  
  //clean up
  velAnaParams.pDataObs->unInit();
  SAFE_DELETE(pFWIObj);
  //We use auto_ptr now, no need to delete those ptrs.
  return 0;
}

