#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "modelingFD2d_op.h"
#include "modelingFDgpu.h"
#include "modelingOpOneway.h"
#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"

#include "param_entry.h"
#include "myOMP.h"

#include "FWI_FD.h"
#include "WEMVA.h"


#include "stackTraceC.h"
#include "FWI_objFunc.h"
#include "solverNonlinearSearch.h"
#include "commonOps.h"

#include "FWI_FD.h"
#include "FWI_FD_beamFocus.h"
#include "FWI_FD_DSO.h"
#include "WEMVA.h"
///Read parameters
#include "nonLinSolver.param.h"
#include "wavePropagate.param.h"
#include "velAna.param.h"
//#include "../misc.h"

/**
 * For beam focusing, Wave Equation Tomography, Just do the 
 * non-linear line search 
 */

//Reflection case

using namespace Yang;
using namespace std;

   
//Doing beam wave-equation tomography for reflection case.
//Use cuda Wave modeling engine

/// The input of the program
/// initial velocity model vel
/// observed data data_obs along with its geometry

int main (int argc, char *argv[])
{
   initCallStackTrace();

   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"vel"));
   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
   waveletObj->loadWavelet();
   velObj->loadVel();
   
   modelingOp *modelingObj = NULL;
   //Read in the parameters for the modeling Operator
   FDModelingParams params;
   velAnaParam velAnaParams;
   WEMVA_Param WEMVAparam;

   bool bUseOneWay = pars->get_bool("bUseOneway",false);
   params.bUseOnewayPropagator = bUseOneWay;
	 VERIFY(bUseOneWay == true); //only allows one-way
	 //might use one-way
   if (bUseOneWay) {
     modelingObj = new modelingOpOneway();
     //Read in the oneway file tags
     string sz_csou = "csou";
     VERIFY(tagExist(sz_csou));
     params.oper = pars->get_string("OneWayOper","foufd");
     params.pdip = pars->get_string("pdip","45");
     params.nws  = pars->get_int("nws");
     params.crecFileName = pars->get_string(csz_tag_dataObs); //!assume we use the same data every time
     params.csouFileName = pars->get_string("csou");
     params.nxPadOneway  = pars->get_int(csz_npad);
     velAnaParams.dObsTag = string(csz_tag_dataObs); //For Imaging, observed data tag
   } else {
   }

   params.pVel = velObj.get();
   params.pWavelet = waveletObj.get();

   params.modelGridShape = velObj->domain();
   params.velModelShape = velObj->domain();
   params.finalImageShape = params.velModelShape;
   params.ndim = params.velModelShape.get_ndim();
   
   read_twoway_prop_param(&params,pars);
   
   //calc time parameters
   float maxFreq,vmax,vmin;
   //Read in the peak/center freq of the seismic wavelet
   maxFreq = waveletObj->get_float("peakFreq");
   waveletObj->setPeakFreq(maxFreq);
   vmax = velObj->maxval();
   vmin = velObj->minval();
   calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.velModelShape);
   
   read_migration_param(&params,pars);
  if (!bUseOneWay) {
  }

  WEMVAparam.off_img_param = params.mOffImgParams;
  WEMVAparam.imageShape    = params.finalImageShape;
  
  read_velAna_param(&velAnaParams,pars);
  read_WEMVA_param(&WEMVAparam,pars);   
  
  
  WEMVA_FD* pWEMVAobj = NULL;
  modelingOp* pModeling = modelingObj;
  pWEMVAobj = create_WEMVA_obj(&velAnaParams, &params, pModeling, &WEMVAparam,pars);

  //After the initialization of params, init the non-linear solver with observed data
  objFunc* pObjFunc = new WEMVA_objFunc(pWEMVAobj);

  solverParam solvPar;
  read_nonLinSolver_par(&solvPar,pars);

  //Create the solver
  gVecComponent initSlowModelVec, finalModelVec;
  
  params.pVel->copyTo(initSlowModelVec);
  initSlowModelVec.reciprocal();
  params.pVel->allocMemoryTo(finalModelVec);
  solverNonlinearSearch nonLinSolver(pObjFunc, &initSlowModelVec,&solvPar);
  
  
  
  stdVecFloat finalErrorVec; int finalIter;
  //no need to generate gradient mask, beacause the deltaS is generated Windowed by the apertures of each shot respectively.
  vector<gVecComponent*> modelsVec, gradsVec, datsVec;
  nonLinSolver.solve( finalModelVec, finalErrorVec, finalIter, 
                     &modelsVec, &gradsVec, &datsVec);
  
  LOG(m4);
  //Save result
  finalModelVec.dump("out");
  finalModelVec.unInit();
  
  //dump error
  int iSize = finalErrorVec.size(); 
  PRV(iSize);
  cubeData<float> finalErrCube(iSize);
  myMemcpy(finalErrCube.pval,&finalErrorVec[0],iSize);
  finalErrCube.dump("error");
  LOG(m5);
  //dump models and grads
  PRV(modelsVec.size()); PRV(gradsVec.size());
  dumpVecModels(modelsVec,"mods");
  dumpVecModels(gradsVec,"grads");
  dumpVecModels(datsVec,"dats");
  LOG(deleteVecModels);
  deleteVecPointers(modelsVec);
  deleteVecPointers(gradsVec);
  LOG(dump final image);
  //Dump the final data
  const cubeData<float>& finalImage = pObjFunc->getCalcData();
  finalImage.dump("fImage");
  LOG(final clean up);
  //clean up
  
  SAFE_DELETE(pWEMVAobj);
  SAFE_DELETE(pObjFunc);
  return 0;
}

